#ifndef PROCESS_EVENTS_H__
#define PROCESS_EVENTS_H__

# include "libR.h"

// begin cdef

void process_events(void);

void polled_events(void);

int peek_event(void);

// end cdef

#endif /* end of include guard: PROCESS_EVENTS_H__ */
